-- If you use this code please insert my copyrights.
-- (c) 2014 V10 ( http://v10.name ) 

require "beltstorage_stack"
require "utils"

DIRECTION_LEFT_TO_RIGHT = 1
DIRECTION_RIGHT_TO_LEFT = 2
DIRECTION_UP_TO_DOWN = 3
DIRECTION_DOWN_TO_UP = 4

BeltStorageDirectionNames = {
  ["left-to-right"] = DIRECTION_LEFT_TO_RIGHT,
  ["right-to-left"] = DIRECTION_RIGHT_TO_LEFT,
  ["up-to-down"] = DIRECTION_UP_TO_DOWN,
  ["down-to-up"] = DIRECTION_DOWN_TO_UP
}

local FIRST_ITEMSIDE = 1
local SECOND_ITEMSIDE = 2

local ITEMSEARCH_OFFSET1 = 0.70
local ITEMSEARCH_OFFSET2 = 0.40
local ITEMSEARCH_OFFSET3 = 0.35
local ITEMSPAWN_OFFSET1 = 0.55
local ITEMSPAWN_OFFSET2 = 0.25

local BeltStorageAreaOffsets = {
  {
  	{{x=-ITEMSEARCH_OFFSET1, y=-ITEMSEARCH_OFFSET2}, {x=0, y=-ITEMSEARCH_OFFSET3}},
  	{{x=-ITEMSEARCH_OFFSET1, y=ITEMSEARCH_OFFSET3}, {x=0, y=ITEMSEARCH_OFFSET2}},
  },
  {
  	{{x=0, y=-ITEMSEARCH_OFFSET2}, {x=ITEMSEARCH_OFFSET1, y=-ITEMSEARCH_OFFSET3}},
  	{{x=0, y=ITEMSEARCH_OFFSET3}, {x=ITEMSEARCH_OFFSET1, y=ITEMSEARCH_OFFSET2}},
  },
  {
  	{{x=-ITEMSEARCH_OFFSET2, y=-ITEMSEARCH_OFFSET1}, {x=-ITEMSEARCH_OFFSET3, y=0}},
  	{{x=ITEMSEARCH_OFFSET3, y=-ITEMSEARCH_OFFSET1}, {x=ITEMSEARCH_OFFSET2, y=0}},
  },
  {
  	{{x=-ITEMSEARCH_OFFSET2, y=0}, {x=-ITEMSEARCH_OFFSET3, y=ITEMSEARCH_OFFSET1}},
  	{{x=ITEMSEARCH_OFFSET3, y=0},{x=ITEMSEARCH_OFFSET2, y=ITEMSEARCH_OFFSET1}},
  }
}

local BeltStorageSpawnOffsets = {
  {
     {x=ITEMSPAWN_OFFSET1, y=-ITEMSPAWN_OFFSET2},
     {x=ITEMSPAWN_OFFSET1, y=ITEMSPAWN_OFFSET2}
  },
  {
     {x=-ITEMSPAWN_OFFSET1, y=-ITEMSPAWN_OFFSET2},
     {x=-ITEMSPAWN_OFFSET1, y=ITEMSPAWN_OFFSET2}
  },
  {
     {x=-ITEMSPAWN_OFFSET2, y=ITEMSPAWN_OFFSET1},
     {x=ITEMSPAWN_OFFSET2, y=ITEMSPAWN_OFFSET1}
  },
  {
     {x=-ITEMSPAWN_OFFSET2, y=-ITEMSPAWN_OFFSET1},
     {x=ITEMSPAWN_OFFSET2, y=-ITEMSPAWN_OFFSET1}
  }
}

local BeltStorageEntityNames = {}
function BeltStorage_AddEntityName(name, direction) BeltStorageEntityNames[name] = direction end
function IsBeltStorage(entity) return BeltStorageEntityNames[entity.name] end

function BeltStorage_new(entity, direction)
  res = {
    entity = entity,
    direction = direction,
    order = {StorageStack_new(), StorageStack_new()},
    scan_areas = {
      Util_calc_bounds(entity.position, BeltStorageAreaOffsets[direction][FIRST_ITEMSIDE]),
      Util_calc_bounds(entity.position, BeltStorageAreaOffsets[direction][SECOND_ITEMSIDE])
    },
    spawn_positions = {
      Util_calc_position(entity.position, BeltStorageSpawnOffsets[direction][FIRST_ITEMSIDE]),
      Util_calc_position(entity.position, BeltStorageSpawnOffsets[direction][SECOND_ITEMSIDE])
    }
  }
  return res
end

function BeltStorage_add(storage)
  table.insert(glob.belt_storages, storage)
end

local function BeltStorage_recalculate_items(storage)

end

function BeltStorage_on_game_init(storage)
  BeltStorage_recalculate_items(storage)
end

local function BeltStorage_scan_items(storage, side)
  local items = game.findentitiesfiltered{name="item-on-ground", area=storage.scan_areas[side]}
  local item_ent = items[1]
  if item_ent then
    local stack = item_ent.stack
    local spawnpos = storage.spawn_positions[side]
    if game.canplaceentity{name="item-on-ground", position = spawnpos, stack=stack} then
      -- adding itemname to storage_stack itemnames for fix bug when item spawned in another side because max_items == 0
      --StorageStack_add_item_name(storage.order[side], stack.name)  --no using function for optimize
      local itemnames = storage.order[side].itemnames
      local itemname = stack.name
      if not itemnames[itemname] then
        itemnames[itemname] = true
        table.insert(itemnames, itemname)
      end

      -- bug: factorio api have bug: after teleport item not moved on belt, so jest spawn new and destroy old
      --item_ent.teleport(spawnpos)
      game.createentity{position=spawnpos, name="item-on-ground", stack=stack}
      item_ent.destroy()
      return 2
    else
      local entity = storage.entity
      if entity.caninsert(stack) then
        entity.insert(stack)
        StorageStack_insert(storage.order[side], stack.name)
        item_ent.destroy()
        return 1
      end
      return -1
    end
  end
  return 0
end

local function BeltStorage_get_next_item(storage, storage_stack, inventory)
  local itemname = StorageStack_getback(storage_stack)
  if itemname == nil then
    -- stack is empty
    local itemnames = storage_stack.itemnames
    local max_items = table.maxn(itemnames) --get items names count
    if max_items == 1 then
      -- this semi-line used only for one item type - getting last
      itemname = StorageStack_getlast(storage_stack)
      if itemname and inventory.getitemcount(itemname) > 0 then
        -- returning last item name, allowing put to stack
        return itemname, false
      end
      -- no items for spawn - return none
    elseif max_items == 0 then
      -- no any items picked up - try spawn anything from chest
      for name, count in pairs(inventory.getcontents()) do
        return name, false
      end
      -- no items inside chest - return none
      return nil, false
    else
      -- semi-line is have more than one item type
      -- get any
      local i = 1
      local lastany = storage_stack.lastany_index
      for i = 1, max_items do
        --itemname = StorageStack_getany(storage_stack)
 	itemname = itemnames[lastany]
        lastany = lastany + 1
        if lastany > max_items then lastany = 1 end

        if itemname and inventory.getitemcount(itemname) > 0 then
          storage_stack.lastany_index = lastany
          return itemname, false
        end
      end
      storage_stack.lastany_index = lastany

      -- no items for spawn - return none 
    end
    -- no items for spawn - return none
  else
    -- have item in stack
    if inventory.getitemcount(itemname) > 0 then
      -- inventory have item - returning it
      return itemname, true
    end

    local max_items = table.maxn(storage_stack.itemnames) --get items names count
    if max_items == 1 then
      -- this semi-line used only for one item type - no items in inventory - return none
      return nil, false
    end

    -- 50 try count for try get another item from stack
    local i = 1
    for i = 1, 50 do    
      itemname = StorageStack_getback(storage_stack)
      if itemname and inventory.getitemcount(itemname) > 0 then
        -- found inventory item in stack - return it
        return itemname, true
      elseif itemname ~= nil then
        -- stack is empy, all other checks in next frame - return none
        return nil, false
      end
    end
    -- all other checks in next frame - return none
  end
  return nil, false
end

local function BeltStorage_spawn_item(storage, side)
  local inventory = storage.entity.getinventory(defines.inventory.chest)
  local storage_stack = storage.order[side]
  local itemname, returnable = BeltStorage_get_next_item(storage, storage_stack, inventory)
  if itemname then
    local stack = {name=itemname, count=1}
    local spawnpos = storage.spawn_positions[side]
    if game.canplaceentity{name="item-on-ground", position = spawnpos, stack=stack} then
      game.createentity{name="item-on-ground", position = spawnpos, stack=stack}
      inventory.remove(stack)
      return 2
    elseif returnable then
      StorageStack_return(storage_stack, itemname)
      return 1
    end
  end
  return 0
end

function BeltStorage_frame(storage)
  local firstspawned = BeltStorage_scan_items(storage, FIRST_ITEMSIDE)
  local secondspawned = BeltStorage_scan_items(storage, SECOND_ITEMSIDE)
  if firstspawned == 0 then
    firstspawned = BeltStorage_spawn_item(storage, FIRST_ITEMSIDE)
  end
  if secondspawned == 0 then
    secondspawned = BeltStorage_spawn_item(storage, SECOND_ITEMSIDE)
  end
end

function BeltStorage_Dump(storage)
  if storage.direction == DIRECTION_LEFT_TO_RIGHT then
    game.player.print("LEFT_TO_RIGHT")
  elseif storage.direction == DIRECTION_RIGHT_TO_LEFT then
    game.player.print("RIGHT_TO_LEFT")
  elseif storage.direction == DIRECTION_UP_TO_DOWN then
    game.player.print("UP_TO_DOWN")
  else
    game.player.print("DOWN_TO_UP")
  end
 
  for side = 1, 2 do
    local storage_stack = storage.order[side]
    local max_items = table.maxn(storage_stack.itemnames)

    if side == FIRST_ITEMSIDE then
       game.player.print("FIRST SIDE STACK:")
    else
       game.player.print("SECOND SIDE STACK:")
    end
    for i = storage_stack.first, storage_stack.last do
      if storage_stack[i] then
        game.player.print(string.format("%s: %d", storage_stack[i].itemname, storage_stack[i].count))
      else
        game.player.print(string.format("Warning: cannot read item stack at index %d", i))
      end
    end

    game.player.print(string.format("ITEMS_COUNT = %d", max_items))
    local items = ""
    for i = 1, max_items do
      if items == "" then
        items = storage_stack.itemnames[i]
      else
        items = items .. "; " .. storage_stack.itemnames[i]
      end
    end
    game.player.print("ITEMS: " .. items )
  end
end
